# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['eesyldap', 'tests']

package_data = \
{'': ['*'],
 'tests': ['attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_boolean.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_date.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_datetime.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_integer.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_json.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_password.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_shadowlastchange.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py',
           'attributes/test_text.py']}

install_requires = \
['beaker>=1.11.0,<2.0.0',
 'passlib>=1.7.2,<2.0.0',
 'python-dateutil>=2.8.1,<3.0.0',
 'python-ldap',
 'pytz>=2019.3,<2020.0',
 'unidecode>=1.1.1,<2.0.0']

setup_kwargs = {
    'name': 'eesyldap',
    'version': '0.3.0',
    'description': 'Simple LDAP Objects mapper',
    'long_description': '# EesyLDAP\n\nThis library offer an easy way to manipulate LDAP objects.\n\n## Getting Started\n\n    git clone https://gitlab.easter-eggs.com/brenard/eesyldap.git\n    cd eesyldap\n    python3 -m venv venv\n    source venv/bin/activate\n    pip install poetry\n    poetry install\n\nSee ```doc/quickstart.md``` to learn how to use this library.\n\n## Run tests\n\nTo run tests, you need to install eesyldap (see _Getting Started_ section) and run the following command being in the root folder of the project : `pytest --disable-warnings --verbose`\n\n## License\nEesyLDAP is free software; you can redistribute it and/or modify it under the terms of the GNU Affero General Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option) any later version.\n\nEesyLDAP is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Affero General Public License for more details.\n\nYou should have received a copy of the GNU Affero General Public License along with this program.  If not, see <http://www.gnu.org/licenses/>.\n',
    'author': 'Benjamin Renard',
    'author_email': 'brenard@easter-eggs.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.easter-eggs.com/brenard/eesyldap',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
